@@ -1,5 +1,6 @@
 #!/bin/bash
 
+set -x
 set -o errexit
 set -o nounset
 set -o pipefail
@@ -8,11 +9,15 @@ PROJECT_ROOT=$(cd $(dirname "$0")/.. ; pwd)
 CODEGEN_PKG=${PROJECT_ROOT}/vendor/k8s.io/kube-openapi
 VERSION="v1alpha1"
 
-go run ${CODEGEN_PKG}/cmd/openapi-gen/openapi-gen.go \
+export GO111MODULE=off
+go build -o dist/openapi-gen ${CODEGEN_PKG}/cmd/openapi-gen
+
+./dist/openapi-gen \
   --go-header-file ${PROJECT_ROOT}/hack/custom-boilerplate.go.txt \
   --input-dirs github.com/argoproj/argo-cd/pkg/apis/application/${VERSION} \
   --output-package github.com/argoproj/argo-cd/pkg/apis/application/${VERSION} \
   --report-filename pkg/apis/api-rules/violation_exceptions.list \
   $@
 
-go run ./hack/gen-crd-spec/main.go
+go build -o ./dist/gen-crd-spec ${PROJECT_ROOT}/hack/gen-crd-spec
+./dist/gen-crd-spec
